from BERT.Base import BERT
import datetime
import numpy
import random
import time
import torch
from torch.optim import AdamW
from transformers import get_linear_schedule_with_warmup

def chrono(f):
    def wrapped(*args, **kwargs):
        t0 = time.time()
        f(*args, **kwargs)
        duration = datetime.timedelta(seconds=round(time.time() - t0))
        print(f"\n  {f.__name__} took: {duration}")
    return wrapped

class Trainer(BERT):
    def __init__(self, root_path, labeled_data, epochs=4):
        self.epochs = epochs
        BERT.__init__(self, root_path, train_on=labeled_data.unique)
        self._init_utils(labeled_data.load(self))

    def _init_utils(self, data_loader):
        self.optimizer = AdamW(
                self.model.parameters(),
                lr = 2e-5, # args.learning_rate - default is 5e-5
                )
        self.data_loader = data_loader
        self.scheduler = get_linear_schedule_with_warmup(
                self.optimizer,
                num_warmup_steps = 0, # Default value in run_glue.py
                num_training_steps = self.epochs * len(data_loader))

    def __call__(self, seed_value=42):
        random.seed(seed_value)
        numpy.random.seed(seed_value)
        torch.manual_seed(seed_value)
        torch.cuda.manual_seed_all(seed_value)
        for e in range(self.epochs):
            self.epoch(e)
        self.save()
        print("\nTraining complete!")

    @chrono
    def epoch(self, epoch):
        self._start_epoch(epoch)
        self.model.train()
        total_loss = sum([self.learn_on(*self.import_data(batch))
                          for batch in self.data_loader])
        avg_train_loss = total_loss / len(self.data_loader)
        print("\n  Average training loss: {0:.2f}".format(avg_train_loss))
        return avg_train_loss

    def learn_on(self, input_ids, input_mask, labels):
        self.model.zero_grad()
        outputs = self.model(input_ids,
                        token_type_ids=None,
                        attention_mask=input_mask,
                        labels=labels)
        loss = outputs[0]
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)
        self.optimizer.step()
        self.scheduler.step()
        return loss.item()

    def _start_epoch(self, epoch):
        print(f'\n======== Epoch {epoch+1} / {self.epochs} ========')
        print('Training...')
